#include <sys/stat.h>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <dlfcn.h>

#include "gauss.hh"
#include "system.hh"
#include "log.hh"

namespace Gauss {
  int debug = Log::DEBUG;
  string exe_dir;     // directory with gausssian executables
  string exe_name = "g98"; // g98 executable name
  int fail_count = 0;

  // potential correction
  extern "C" {
    typedef double  (*pcorr_t) (const int&, const double [], 
				const double [], const int [], double []);
  }
  pcorr_t _pcorr = 0;
  bool  _ispcorr = false;
  double* _rpar = 0; // array of real parameters for potential correction
  int*    _ipar = 0; // array of integer parameters for potential correction

  // system specification
  vector<Atom> cluster;
  int multiplicity;
  int charge;

  // methods
  vector<Method*> Method::avail_methods;
  vector<Interpolation> interpolation_pool; // interpolations to try
  vector<Methp>         minimization_pool; // to look for minimum
  Methp opt_method; // method for optimization

  // functions
  void test_stream (const ifstream&, const char*) throw(File_Err);
}

double Gauss::pcorr(const int& n, const double r [], double f [])
{
  if(_ispcorr)
    return _pcorr(n, r, _rpar, _ipar, f);
  cout << "Gauss::pcorr: potential is not defined\n";
  return 0.;    
}

bool Gauss::ispcorr () { return _ispcorr; }

void Gauss::test_stream (const ifstream& from, const char* funame) 
  throw(File_Err)
{
  if (!from) // stream input has failed
    if (from.eof()) { // end of file
      if (debug >= Log::WARN)
	Log::out << funame << "end of file encountered\n";
      throw EOF_Err();
    }
    else {
      Log::out << funame << "Oops!!! format error\n";
      throw Form_Err();
    }
    
}

/*****************************************************
 ******************* Atom class **********************
 *****************************************************/

ostream& Gauss::operator<< (ostream& to, const Atom& a)
{
  to << a.num;
  for (int i = 0; i < 3; ++i)
    to << "   " << a.pos[i];
  to << "\n";
  return to;
}

/*****************************************************
 *************** Buffstream class ********************
 *****************************************************/

Gauss::Buffstream::Buffstream(const string& fname) throw(Open_Err)
  : from(fname.c_str())
{
  static const char* funame = "Gauss::Buffstream::Buffstream: ";

  if(!from) {
    Log::out << funame << "Oops!!! cannot open " 
	    << fname << " file\n";
    throw Open_Err();
  }
}

const string& Gauss::Buffstream::operator[] (int index) 
  throw(Err)
{
  static const char* funame = "Gauss::Buffstream::operator[]: ";

  if (index < 0) {
    Log::out << funame << "Oops!!! out of range\n";
    throw Range_Err();
  }

  string tmps;
  while(buff.size() <= index) {
    from >> tmps;
    test_stream(from, funame);
    buff.push_back(tmps);
  }
  return buff[index];
}
    
void Gauss::Buffstream::skip (int index) throw(Err)
{
  static const char* funame = "Gauss::Buffstream::skip: ";
  
  if (index < 0) {
    Log::out << funame << "Oops!!! out of range\n";
    throw Range_Err();
  }

  int bs = buff.size();
  if (index < bs)
    buff.erase(buff.begin(), buff.begin()+index);
  else {
    buff.clear();
    string tmps;
    for (; bs < index; ++bs) { 
      from >> tmps;
      test_stream(from, funame);
    }
  }
}

Gauss::Buffstream& Gauss::Buffstream::operator>> (double& dv) 
  throw(File_Err)
{
  static const char* funame = "Gauss::Buffstream::operator>>: ";

  std::istringstream from((*this)[0]);
  from >> dv;
  if(from.fail()) { // check the string
    Log::out << funame << "Oops!!! the string " << buff[0] 
	 <<  " is not a valid representation of double\n";
    throw Form_Err();
  }

  buff.erase(buff.begin()); // clean the buffer

  return *this;
}

Gauss::Buffstream& Gauss::Buffstream::operator>> (string& s) 
  throw(File_Err)
{


  s =(*this)[0];
  buff.erase(buff.begin()); // clean the buffer

  return *this;
}

/***********************************************************
 ******************** Methp classes **********************
 ***********************************************************/


Gauss::Method* Gauss::Methp::operator-> () const throw(Init_Err) 
{
  static const char* funame = "Gauss::Methp::operator->: ";
  if (mp)
    return mp;
  else {
    Log::out << funame << "Oops!!! not initialized\n";
    throw Init_Err();
  }
}

istream& Gauss::operator>> (istream& from, Methp& meth)
{
  return from >> meth.id;
}

ostream& Gauss::operator<< (ostream& to, const Methp& meth)
{
  return to << meth.id;
}

/***********************************************************
 ******************** Pattern classes **********************
 ***********************************************************/

Gauss::Pattern::Pattern (ifstream& from) throw(Err)
{
  static const char* funame = "Gauss::Pattern::Pattern: ";


  string stemp;

  string separator;

  from >> separator;
  while (from >> stemp) {
    if (stemp == separator)
      break;
    keys.push_back(stemp);
  }

  test_stream(from, funame);
  
  if (!keys.size()) {
    Log::out << funame 
	 << "Oops!!! there should be at least one search key in the pattern\n";
    throw Init_Err();
  }
}

bool Gauss::Pattern::test (Buffstream& sb) const throw()
{
  try {
    for (int i = 0; i < keys.size(); ++i)
      if (sb[i] != keys[i])
	return false;
  }// try
  catch(EOF_Err) {
    return false;
  }
  // pattern matching succeeded
  sb.skip(keys.size());
  return true;
}

Gauss::Value_pattern::Value_pattern (ifstream& from) throw(Err)
  : Pattern(from)
{
  static const char* funame = "Gauss::Value_pattern::Value_pattern: ";
  from >> _weight;
  test_stream(from, funame);
}

Gauss::Message_pattern::Message_pattern (ifstream& from) throw(Err)
  : Pattern(from)
{
  static const char* funame = "Gauss::Message_pattern::Message_pattern: ";

  string separator;
  string stemp;

  // read message
  bool is_first = true;
  from >> separator;
  while (from >> stemp) {
    if (stemp == separator)
      break;
    if(is_first) {
      is_first = false;
      _message = stemp;
    }
    else
      _message += " " + stemp;
  }
  test_stream(from, funame);
}

Gauss::Fail_pattern::Fail_pattern (ifstream& from) throw(Err)
  : Message_pattern(from)
{
  static const char* funame = "Gauss::Fail_pattern::Fail_pattern: ";

  string separator;
  string stemp;

  // read methods
  from >> separator;
  while (from >> stemp) {
    if (stemp == separator)
      break;
    methods.push_back(stemp);
  }

  test_stream(from, funame);

}

Gauss::Methp& Gauss::Fail_pattern::method(int i) throw(Range_Err) 
{ 
  static const char* funame = "Gauss::Fail_pattern::method: ";

  if (i < 0 || i >= methods.size()) {
    Log::out << funame << "Oops!!! out range\n";
    throw Range_Err();
  }
  return methods[i];
}

/****************************************************************
 ********************* Method class *****************************
 ****************************************************************/

void Gauss::Method::clean_all () throw()
{
  for (int i = 0; i < avail_methods.size(); ++i)
    avail_methods[i]->init();
}

void Gauss::Method::find_method (Methp& mid) throw(Find_Err)
{
  static const char* funame = "Gauss::Method::find_method: ";
  for (int i = 0; i < avail_methods.size(); ++i)
    if (avail_methods[i]->name() == mid.id) {
      mid.mp = avail_methods[i];
      return;
    }

  Log::out << funame << "Oops!!! method " << mid 
       << " not found. Check the g98 initialization file\n";

  throw Find_Err();
}

void Gauss::Method::check (vector<string>& mstack) throw(Init_Err)
{
  static const char* funame = "Gauss::Method::check: ";

  if (is_checked)
    return;  

  string stemp;
  int itemp = 0;

  // check if the current method falls into an infinite loop
  for (int i = 0; i < mstack.size(); ++i)
    if (mstack[i] == id) {
      Log::out << funame << "Oops!!! the methods form an infinite loop:\n";
      for (int j = i; j < mstack.size(); ++j) {
	if (stemp.size() + mstack[j].size() + 4 - itemp > 60) {
	  stemp += "\n";
	  itemp = stemp.size();
	}
	stemp += mstack[j] + " -> ";
      }
      if (stemp.size() + id.size() - itemp > 60)
	stemp += "\n";
      stemp += id;
      Log::out << stemp << "\n";

      throw Init_Err();
    }

	
  // check derived methods;
  mstack.push_back(id);

  try {
    // check initial guess methods
    for (int i = 0; i < init_guess_methods.size(); ++i) {
      find_method(init_guess_methods[i]);
      init_guess_methods[i]->check(mstack);
    }

    // check default fail-safe methods
    for (int i = 0; i < default_fail_safe_methods.size(); ++i) {
      find_method(default_fail_safe_methods[i]);
      default_fail_safe_methods[i]->check(mstack);
    }

    // check failure-specific fail-safe methods
    for (int i = 0; i < fail_patt.size(); ++i)
      for (int j = 0; j < fail_patt[i].size(); ++j) {
	find_method(fail_patt[i].method(j));
	fail_patt[i].method(j)->check(mstack);
      }
  }
  catch (Find_Err) {
    Log::out << funame << "Oops!!! cannot find one of the method " << id 
	 << " immediate derivatives among the available methods\n";
    throw Init_Err();
  }

  is_checked = true;
  mstack.pop_back();
}

void Gauss::Method::check_all () throw(Init_Err)
{
  vector<string> mstack;
  mstack.reserve(avail_methods.size());
  for (int i = 0; i < avail_methods.size(); ++i)
    avail_methods[i]->check(mstack);
}

Gauss::Method::Method (ifstream& from) throw(Err)
  : is_init_guess_required(false), is_init_guess_min(false), 
    is_init_guess_mix(false), mcorr(0), ref_en(0.), _abs_en(0.),
    is_checked(false), _status(INIT), failsafe(0) 
{
  static const char* funame = "Gauss::Method::Method: ";

  static const string command_key    = "Command";
  static const string init_guess_key = "InitialGuessMethods";
  static const string init_req_key   = "InitialGuessRequired";
  static const string init_min_key   = "InitialGuessMinimum";
  static const string init_mix_key   = "InitialGuessMix";
  static const string ener_patt_key  = "EnergySearchPattern";
  static const string fail_patt_key  = "FailSearchPattern";
  static const string run_patt_key   = "RerunPattern";
  static const string default_key    = "DefaultFailSafeMethods";
  static const string mcorr_key      = "MultiplicityCorrection";
  static const string refen_key      = "ReferenceEnergy";
  static const string finish_key     = "EndMethod";              
  static const string mem_key        = "Memory";

  static const char* all_keys [] = {
    command_key.c_str(),   
    init_guess_key.c_str(),
    init_req_key.c_str(),  
    init_min_key.c_str(),  
    init_mix_key.c_str(),  
    ener_patt_key.c_str(), 
    fail_patt_key.c_str(), 
    run_patt_key.c_str(),  
    default_key.c_str(),   
    mcorr_key.c_str(),     
    refen_key.c_str(),     
    finish_key.c_str(),    
    mem_key.c_str(),       
    0
  };

  string stemp;
  int    itemp;


  string separator;

  bool is_init_err = false;
  from >> id; // method name

  for (int i = 0; i < avail_methods.size(); ++i)
    if (avail_methods[i]->id == id) {
      Log::out << funame << "method " << id << " is allready available\n";
      is_init_err = true;
    }

  while (from >> stemp) {// input cycle

    if (stemp == command_key) {
      if (command.size()) {
	Log::out << funame 
	     << "there should be exactly one " << command_key
	     << " keyword for " << id << " method\n";
	is_init_err = true;
      }
      from >> separator;
      itemp = 0; // last new line position
      while (from >> stemp) {
	if (stemp == separator)
	  break;
	if (command.size() + stemp.size() + 1 - itemp > 70) {
	  command += "\n";
	  itemp = command.size();
	}
	if (command.size() != itemp)
	  command += " " + stemp;
	else
	  command += stemp;
      }
      stemp = "units=bohr";
      if (command.size() + stemp.size() + 1 - itemp > 70) {
	command += "\n";
	itemp = command.size();
      }
      if (command.size() != itemp)
	command += " " + stemp;
      else
	command += stemp;

      command += "\n";
    }

    else if (stemp == init_guess_key) {
      if (init_guess_methods.size()) {
	Log::out << funame 
	     << "there should be not more than one " << init_guess_key
	     << " keyword in " << id << " method\n";
	is_init_err = true;
      }
      from >> separator;
      while (from >> stemp) {
	if (stemp == separator)
	  break;
	init_guess_methods.push_back(stemp);
      }	
    }

    else if (stemp == init_req_key) {
      from >> is_init_guess_required;
    }

    else if (stemp == init_min_key) {
      from >> is_init_guess_min;
    }

    else if (stemp == init_mix_key) {
      from >> is_init_guess_mix;
    }
    
    else if (stemp == mem_key) {
      from >> memory;
    }

    else if (stemp == ener_patt_key) {
      Value_pattern vtemp(from);
      ener_patt.push_back(vtemp);
    }

    else if (stemp == fail_patt_key) {
      Fail_pattern ftemp(from);
      fail_patt.push_back(ftemp);
    }

    else if (stemp == run_patt_key) {
      Message_pattern ftemp(from);
      run_patt.push_back(ftemp);
    }

    else if (stemp == default_key) {
      if (default_fail_safe_methods.size()) {
	Log::out << funame 
	     << "there should be not more than one " << default_key
	     << " keyword in " << id << " method\n";
	is_init_err = true;
      }
      from >> separator;
      while (from >> stemp) {
	if (stemp == separator)
	  break;
	default_fail_safe_methods.push_back(stemp);
      }
    }

    else if (stemp == mcorr_key) {
      from >> mcorr;
    }

    else if (stemp == refen_key) {
      from >> ref_en;
    }

    else if (stemp == finish_key) {
      break;
    }

    else {
      Log::out << funame << "unknown keyword " << stemp 
	   << " in " << id << " method\n\navailable keys:\n";
      for(int i = 0; all_keys[i] != 0; ++i)
	Log::out << all_keys[i] << "\n";
      is_init_err = true;
      break;
    }
  }// input cycle

  test_stream(from, funame); // test stream state

  // test that everything which should be initialized is initialized
  if (!command.size()) {
    Log::out << funame 
	 << "there should be exactly one keyword "
	 << command_key << " in " << id << " method\n";
    is_init_err = true;
  }

  if (!init_guess_methods.size() && is_init_guess_required) {
    Log::out << funame 
	 << "there should be at least one initial guess method in "
	 << id << " method\n";
    is_init_err = true;
  }

  if (!ener_patt.size()) {
    Log::out << funame 
	 << "there should be at least one energy pattern in "
	 << id << " method\n";
    is_init_err = true;
  }

  if(is_init_err)
    throw Init_Err();

  //add the current method to the methods pool
  avail_methods.push_back(this);
}

void Gauss::Method::backup() throw(Init_Err)
{
  static const char funame[] = "Gauss::Method::backup: ";

  if(_status != SUCCESS) {
    Log::out << funame << "method " << id << " status is not SUCCESS\n";
    throw Init_Err();
  }
  file_copy((id + ".chk").c_str(), (id + ".chk.back").c_str());
  file_copy((id + ".log").c_str(), (id + ".log.back").c_str());
  file_copy((id + ".fchk").c_str(),(id + ".fchk.back").c_str());
  backup_ener = _abs_en;
}

void Gauss::Method::restore() throw(Init_Err)
{
  static const char funame[] = "Gauss::Method::restore: ";

  if(rename((id + ".chk.back").c_str(), (id + ".chk").c_str()) ||
     rename((id + ".log.back").c_str(), (id + ".log").c_str()) ||
     rename((id + ".fchk.back").c_str(),(id + ".fchk").c_str())) {
    Log::out << funame << "failed\n";
    throw Init_Err();
  }
  _abs_en = backup_ener;
  _status = SUCCESS;
}

/*
void Gauss::Method::clean () throw()
{
  _status = INIT;

  // remove g98 files
  remove((id + ".com").c_str());
  remove((id + ".log").c_str());
  remove((id + ".fchk").c_str());
  remove((id + ".chk").c_str());
  remove((id + ".rwf").c_str());
  remove((id + ".int").c_str());
  remove((id + ".d2e").c_str());

}
*/

void Gauss::Method::save_log (int count) const throw()
{
  string old_name;
  char new_name [100];

  old_name = id + ".com";
  sprintf(new_name, "%s.%i", old_name.c_str(), count);
  rename(old_name.c_str(), new_name);
  
  old_name = id + ".log";
  sprintf(new_name, "%s.%i", old_name.c_str(), count);
  rename(old_name.c_str(), new_name);
  
  old_name = id + ".chk";
  sprintf(new_name, "%s.%i", old_name.c_str(), count);
  rename(old_name.c_str(), new_name);
  
  old_name = "std.err";
  sprintf(new_name, "%s.%s.%i", id.c_str(), old_name.c_str(), count);
  rename(old_name.c_str(), new_name);
  
  old_name = "std.out";
  sprintf(new_name, "%s.%s.%i", id.c_str(), old_name.c_str(), count);
  rename(old_name.c_str(), new_name);

}

double Gauss::Method::rel_energy () const throw(Init_Err)
{
  static const char* funame = "Gauss::Method::rel_energy: ";

  if (_status != SUCCESS) {
    Log::out << funame << "Oops!!! energy is not initialized\n";
    throw Init_Err();
  }
  return _abs_en - ref_en;
}

double Gauss::Method::abs_energy  () const throw(Init_Err)
{
  static const char* funame = "Gauss::Method::abs_energy: ";

  if (_status != SUCCESS) {
    Log::out << funame << "Oops!!! energy is not initialized\n";
    throw Init_Err();
  }
  return _abs_en;
}

void Gauss::Method::read_forces (vector<D3>* force_array) 
  const throw(Err)
{
  static const char* funame = "Gauss::Method::read_forces: ";

  double dtemp;

  string stemp;

  if (force_array && force_array->size() != cluster.size()) {
    Log::out << funame << "Oops!!! forces array size error\n";
    throw Init_Err();
  }

  if (_status != SUCCESS) {
    Log::out << funame << "Oops!!! method " << id 
	    << " was not successfully applied\n";
    throw Init_Err();
  }

  // find forces
  ifstream fchk((id + ".fchk").c_str());
  if(!fchk) {
    Log::out << funame <<  "Oops!!! cannot open fchk file. "
      "Check files *." << fail_count << "\n\n";
    save_log(fail_count++);
    throw Open_Err();
  }
  bool is_force_found = false;
  while(fchk >> stemp)
    if (stemp == "Gradient") { // forces
      is_force_found = true;
      getline(fchk, stemp);
      for (int at = 0; at < cluster.size(); ++at)
	for (int i = 0; i < 3; ++i) {
	  fchk >> dtemp;
	  if (force_array)
	    (*force_array)[at][i] = -dtemp;
	  else
	    cluster[at].force[i]  = -dtemp;
	}
      if(!fchk)	{
	Log::out << funame <<  "Oops!!! reading forces from fchk file failed. "
	  "Check files *." << fail_count << "\n\n";
	save_log(fail_count++);
	throw Form_Err();
      }
      break;
    }
  
  if (!is_force_found) {
    Log::out << funame << "Oops!!! Gradient keyword is not found "
      "in the fchk file. Check files *." << fail_count << "\n";
    save_log(fail_count++);
    throw Find_Err();
  }
}

Gauss::Method* Gauss::Method::apply (int flags) throw()
{

  static const char* funame = "Gauss::Method::apply: ";

  // check if the method was allready applied
  if (_status == SUCCESS)
    return this;
  else if (_status == FAIL)
    if(flags & DIRECT)
      return 0;
    else
      return failsafe;

  _status = FAIL;
  failsafe = 0;

  string stemp; // string  temporary
  double dtemp; // double  temporary


  /********************* BEGIN initial guess **********************/

  bool is_init_guess = true; // use initial guess checkpoint file
  
  if(flags & READ) {// save checkpoint file
    file_copy((id + ".chk").c_str(), (id + ".chk.read").c_str());
  }
  else {// initial guess
    const Method* guess = 0;
    if (is_init_guess_min)
      for (int i = 0; i < init_guess_methods.size(); ++i) {
	const Method* meth = init_guess_methods[i]->apply();
	if (meth && (!guess || meth->abs_energy() < guess->abs_energy()))
	  guess = meth;
      }// init_guess_min
    else
      for (int i = 0; i < init_guess_methods.size(); ++i) {
	if(guess = init_guess_methods[i]->apply());
	break;
      }

    // set initial guess
    if (!guess) {
      if(is_init_guess_required) {
	if (debug >= Log::WARN)
	  Log::out << funame << "Oops!!! initial guess calculation for method "
		   << id << " failed\n";
	if(flags & DIRECT) return 0;
	// default fail-safe methods 
	for(int i = 0; i < default_fail_safe_methods.size(); ++i) {
	  if (debug >= Log::WARN)
	    Log::out << funame <<"trying default fail-safe method " 
		    << default_fail_safe_methods[i]->name()
		    << "\n";
	  if(failsafe = default_fail_safe_methods[i]->apply(flags))
	    return failsafe;
	}
	return 0;
      }
      else
	is_init_guess = false;
    }
    else if(file_copy((guess->name()+".chk").c_str(),(id+".chk").c_str())) {
      if(is_init_guess_required) {
	if(debug >= Log::ERROR)
	Log::out << funame << "Oops!!! copy of initial guess "
		<< guess->name() + ".chk to " 
		<< id + ".chk failed\n";
	if(flags & DIRECT) return 0;
	// default fail-safe methods 
	for (int i = 0; i < default_fail_safe_methods.size(); ++i) {
	  if(debug >= Log::WARN)
	    Log::out << funame <<"trying default fail-safe method " 
		    << default_fail_safe_methods[i]->name()
		    << "\n";
	  if(failsafe = default_fail_safe_methods[i]->apply(flags))
	    return failsafe;
	}
	return 0;
      }
      else
	is_init_guess = false;
    }
  }//initial guess
  /***************** END initial guess ********************/

  /************* BEGIN create g98 input file **************/

  ofstream gout((id + ".com").c_str());
  gout.setf(ios::fixed, ios::floatfield);

  if (!gout) {
    Log::out << funame << "Oops!!! cannot open g98 input file\n";
    return 0;
  }

  // Link 0 (%) commands
  if(memory.size())
    gout << "%Mem=" << memory <<"\n";
  gout << "%Chk=" << id << "\n";
  gout << "%RWF=" << id << "\n";
  gout << "%Int=" << id << "\n";
  gout << "%D2E=" << id << "\n";
  // Route section
  gout << command;
  if (is_init_guess)
    stemp = "guess=read";
  if (is_init_guess_mix)
    stemp += " guess=mix";
  if (flags & FORCE)
    stemp += " force";
  if (stemp.size())
    gout << stemp << "\n";

  // Title section
  gout << "\n" << funame << "\n\n";

  // Molecule specification section
  gout << charge << " " << multiplicity + mcorr << "\n";
  for (int i = 0; i < cluster.size(); ++i)
    gout << cluster[i];
  gout << "\n";

  if (!gout) {
    Log::out << funame << "Oops!!! writting to g98 input file has failed\n";
    return 0;
  }

  gout.close();

  /************* END create g98 input file **************/

  /****************** BEGIN call g98  ********************/
  int exit_code;
  if(exit_code = call_exe(exe_name.c_str(), id.c_str(), (char*) 0)){// g98 fail
    if(debug >= Log::ERROR) {
      Log::out << funame << "Oops!!! g98 method "
	       << id << " failed with exit code " << exit_code << "\n"
	       << "Check files *." << fail_count << "\n\n";
	      save_log(fail_count++);
    }

    try {
      Buffstream login(id + ".log");
      while(1) {
	for (int patt_ind = 0; patt_ind < run_patt.size(); ++patt_ind)
	  if (run_patt[patt_ind].test(login)) {
	    if (debug >= Log::ERROR) {
	      Log::out << funame << run_patt[patt_ind].message() << "\n";
	    }
	    login.close();
	    _status = INIT;
	    if(flags & READ)
	      rename((id + ".chk.read").c_str(), (id + ".chk").c_str());
	    return apply(flags);
	  }
	login.skip();
      }
    }
    catch(Err) {}

    if(flags & DIRECT)
      return 0;
    
      
    // fail-safe methods
    try {
      Buffstream login(id + ".log");
      while(1) {
	for (int patt_ind = 0; patt_ind < fail_patt.size(); ++patt_ind)
	  if (fail_patt[patt_ind].test(login)) {
	    if (debug >= Log::ERROR) {
	      Log::out << funame << fail_patt[patt_ind].message() << "\n";
	    }
	    for (int i = 0; i < fail_patt[patt_ind].size(); ++i) {
	      if (debug >= Log::INFO)
		Log::out << funame <<"trying fail-safe method " 
			<< fail_patt[patt_ind].method(i)->name()
			<< "\n"; 
	      if(failsafe = fail_patt[patt_ind].method(i)->apply(flags))
		return failsafe;
	    }
	    return 0;
	  }
	login.skip();
      }
    }
    catch(EOF_Err) {
      Log::out << funame << "Oops!!! cannot find failure pattern "
	"in g98 log file.\n";
    }
    catch(Open_Err) {
      Log::out << funame << "Oops!!! cannot open g98 log file\n";
    }
    catch(Form_Err) {
      Log::out << funame << "Oops!!! g98 log file formatt error\n";
    }
    Log::out << "Check files *." << fail_count << "\n\n";
    save_log(fail_count++);
      
    // default fail-safe methods 
    for (int i = 0; i < default_fail_safe_methods.size(); ++i) {

      if (debug >= Log::INFO)
	Log::out << funame <<"trying default fail-safe method " 
		<< default_fail_safe_methods[i]->name()
		<< "\n";
      if(failsafe = default_fail_safe_methods[i]->apply(flags))
	return failsafe;
    }
    return 0;

  }// g98 failed

  else {// g98 succeeded

    // convert checkpoint file
    if (call_exe("formchk", (id + ".chk").c_str(),
		 (id + ".fchk").c_str(), (char*) 0)) {
      Log::out << funame << "checkpoint file conversion failure\n";
      return 0;
    }

    // find energies in fchk file
    try {
      Buffstream fchk(id + ".fchk");  // fchk buffered stream
      int   count = 0;
      _abs_en = 0.;
      while(1) {
	bool is_found = false;
	for (int i = 0; i < ener_patt.size(); ++i)
	  if (ener_patt[i].test(fchk)) {// test fchk for pattern
	    fchk >> dtemp;
	    _abs_en += dtemp * ener_patt[i].weight();
	    count++;
	    is_found = true;
	    break;
	  }
	if(!is_found) fchk.skip();
    	// test if energies have been read
	if(count == ener_patt.size()) {
	  _status = SUCCESS;
	  return this;
	}
      }
    }
    catch(EOF_Err) {
      Log::out << funame << "Oops!!! energy pattern is not found "
	"in g98 fchk file\n";
    }
    catch(Open_Err) {
      Log::out << funame << "Oops!!! cannot open g98 fchk file\n";
    }
    catch(Form_Err) {
      Log::out << funame << "Oops!!! g98 fchk file formatt error\n";
    }
    Log::out << "Check files *." << fail_count << "\n\n";
    save_log(fail_count++);
    return 0;

  }// g98 succeeded

  /****************** END call g98 ********************/
  
}// apply()

void Gauss::Method::print_all_info (ostream& to)
{
  int old_precision = to.precision(9);
  for (int i = 0; i < avail_methods.size(); ++i)
    avail_methods[i]->print_info(to);
  to.precision(old_precision);
  to << "\n";
}

void Gauss::Method::print_info (ostream& to)
{
  if (_status == INIT)
    return;

  to << "Method " << id; 
  if (_status == SUCCESS)
    to << ": Energy = " << abs_energy() << "\n";
  else
    to << " failed\n";

}

/************************************************************
 ***************** Interpolation class **********************
 ************************************************************/

Gauss::Interpolation::Interpolation (ifstream& from) throw(Err)
{
  static const char* funame = "Gauss::Interpolation::Interpolation: ";

  string stemp;
  double dtemp;

  string separator;

  bool is_init_err = false;
  from >> separator;
  while (from >> stemp) {
    if (stemp == separator)
      break;
    from >> dtemp;
    methods.push_back(stemp);
    weights.push_back(dtemp);
  }

  test_stream(from, funame);

  if (!methods.size()) {
    Log::out << funame << "there should be at least one method to apply\n";
    is_init_err = true;
  }

  if (is_init_err)
    throw Init_Err();
}	

void Gauss::Interpolation::check () throw(Init_Err)
{
  static const char* funame = "Gauss::Interpolation::check: ";

  bool is_found = true;
  for (int i = 0; i < methods.size(); ++i)
    try {
      Method::find_method(methods[i]);
    }
    catch(Find_Err) {
      Log::out << funame << "method " << methods[i] 
	   << " is not found in among the available methods\n";
      is_found = false;
    }
  if(!is_found)
    throw Init_Err();
}

double Gauss::Interpolation::apply(int flags) const throw(Err)
{
  static const char* funame = "Gauss::Interpolation::apply: ";

  vector<D3> force_array(cluster.size());
  if (flags & FORCE)
    for (int at = 0; at < cluster.size(); ++at)
      for (int j = 0; j < 3; ++j)
	cluster[at].force[j] = 0.;

  double res = 0.;
  const Method* meth;
  for (int i = 0; i < methods.size(); ++i) {
    meth = methods[i]->apply(flags);
    if (meth) {
      res += meth->rel_energy() * weights[i];
      if (flags & FORCE) {
	meth->read_forces(&force_array);
	for (int at = 0; at < cluster.size(); ++at)
	  for (int j = 0; j < 3; ++j)
	    cluster[at].force[j] += force_array[at][j] * weights[i];
      }
    }
    else {
      Log::out << funame << "method " << methods[i]
	      << " failed\n";
      throw Err();
    }
  }
  return res;
}

//main initialization function
void Gauss::init (const string& fname) throw(Err)
{
  static const char* funame = "Gauss::init: ";

  static const string finish_key = "EndInit";
  static const string  inter_key = "Interpolation";
  static const string method_key = "Method";
  static const string    min_key = "MinMethods";
  static const string  multi_key = "Multiplicity";
  static const string charge_key = "Charge";
  static const string  debug_key = "DebugLevel";
  static const string   exec_key = "ExecDir";
  static const string    g98_key = "ExecName";
  static const string   csym_key = "PotentialCorrectionSymbol";
  static const string   clib_key = "PotentialCorrectionLibrary";
  static const string   rpar_key = "PotentialParameterReal";
  static const string   ipar_key = "PotentialParameterInteger";
  static const string    opt_key = "OptimizationMethod";
  static const string    read_key = "OptimizationReadMethod";

  static const char* all_keys [] = {
    finish_key.c_str(),
    inter_key.c_str(),
    method_key.c_str(),
    min_key.c_str(),
    multi_key.c_str(),
    charge_key.c_str(),
    debug_key.c_str(),
    exec_key.c_str(),
    g98_key.c_str(),
    csym_key.c_str(),
    clib_key.c_str(),
    rpar_key.c_str(),
    ipar_key.c_str(),
    opt_key.c_str(),
    read_key.c_str(),
    0
  };

  string token;



  string separator;

  ifstream from(fname.c_str());
  if(!from) {
    std::cout << funame << "cannot open input file " << fname << "\n";
    throw Open_Err();
  }


  bool is_init_err = false;

  bool is_finish = false;
  bool is_multi  = false;
  bool is_charge = false;
  bool is_debug  = false;
  bool is_exec   = false;
  bool is_g98    = false;
  bool is_csym   = false;
  bool is_clib   = false;
  bool is_rpar   = false;
  bool is_ipar   = false;
  bool is_opt    = false;

  string clib, csym; // dynamic library
  while (from >> token) {
    if (token == finish_key) {
      is_finish = true;
      break;
    }
    else if (token == multi_key) {
      from >> multiplicity;
      is_multi = true;
    }
    else if (token == charge_key) {
      from >> charge;
      is_charge = true;
    }
    else if (token == debug_key) {
      from >> debug;
      is_debug = true;
    }
    else if(token == exec_key) {
      from >> exe_dir;
      is_exec = true;
    }
    else if(token == g98_key) {
      from >> exe_name;
      is_g98 = true;
    }
    else if (token == inter_key) {
      try {
	Interpolation iptemp(from);
	interpolation_pool.push_back(iptemp);
      }
      catch(Init_Err) {
	is_init_err = true;
      }
    }
    else if (token == method_key) {
      try {
	new Method(from);
      }
      catch(Init_Err) {
	is_init_err = true;
      }
    }
    else if (token == min_key) {
      from >> separator;
      while (from >> token) {
	if (token == separator)
	  break;
	minimization_pool.push_back(token);
      }
    }
    else if(token == clib_key) {
      from >> clib;
      if(!from) {
	std::cout << funame << "cannot read " << token << "\n";
	is_init_err = true;
      }
      else
	is_clib = true;
    }
    else if(token == csym_key) {
      from >> csym;
      if(!from) {
	std::cout << funame << "cannot read " << token << "\n";
	is_init_err = true;
      }
      else
	is_csym = true;
    }
  // real parameters
    else if(token == rpar_key) {
      if(is_rpar){
	std::cout << funame << "another " << token << " declaration\n";
	is_init_err = true;
      }

      vector<double> data;
      from >> separator;
      while (from >> token) {
        if (token == separator)
          break;
	data.push_back(atof(token.c_str()));
      }
      if(!from) {
	std::cout << funame << "cannot read " << rpar_key << "\n";
	is_init_err = true;
	break;
      }
      _rpar = new double[data.size()];
      for(int i = 0; i < data.size(); ++i)
	_rpar[i] = data[i];
      is_rpar = true;
    }
  // integer parameters
    else if(token == ipar_key) {
      if(is_ipar){
	std::cout << funame << "another " << token << " declaration\n";
	is_init_err = true;
      }

      vector<int> data;
      from >> separator;
      while (from >> token) {
        if (token == separator)
          break;
	data.push_back(atoi(token.c_str()));
      }
      if(!from) {
	std::cout << funame << "cannot read " << ipar_key << "\n";
	is_init_err = true;
	break;
      }
      _ipar = new int[data.size()];
      for(int i = 0; i < data.size(); ++i)
	_ipar[i] = data[i];
      is_ipar = true;
    }
    else if(token == opt_key) {
      from >> opt_method.id;
      is_opt = true;
    }
    else {
      std::cout << funame << "unknown keyword " << token 
	      << "\n\navailable keys:\n";
      for(int i = 0; all_keys[i] != 0; ++i)
	std::cout << all_keys[i] << "\n";

      is_init_err = true;
      break;
    }
  }

  try {
    test_stream(from, funame);
  }
  catch(EOF_Err) {
    std::cout << funame << "end of file encountered\n";
    is_init_err = true;
  }
  catch(Form_Err) {
    std::cout << funame << "format error\n";
    is_init_err = true;
  }

  // check that everything which should be initialized is initialized
  if(!is_finish) {
    std::cout << funame << "input termination key is missing\n";
    is_init_err = true;
  }
  if(!is_multi) {
    std::cout << funame << "multiplicity is not defined\n";
    is_init_err = true;
  }
  if(!is_charge) {
    std::cout << funame << "charge is not defined\n";
    is_init_err = true;
  }
  if(!is_debug) {
    std::cout << funame << "debugging level is not defined\n";
    is_init_err = true;
  }
  if(!is_exec) {
    std::cout << funame << "g98 exec directory is not defined\n";
    is_init_err = true;
  }
  if(!is_g98) {
    std::cout << funame 
	    << "g98 executable name is not defined, using the default: "
	    << exe_name << "\n";
  }

  if (!Method::pool_size()) {
    std::cout << funame 
	 << "there should be at least one method specification available\n";
    is_init_err = true;
  }

  if (!minimization_pool.size() && !interpolation_pool.size()) {
    std::cout << funame 
	 << "there should be either minimization or "
      "interpolation descriptions available\n";
    is_init_err = true;
  }

  // check the methods tree for infinite loops
  try {
    Method::check_all();
  }
  catch(Init_Err) {
    is_init_err = true;
  }

  // check interpolation methods
  for (int i = 0; i < interpolation_pool.size(); ++i)
    try {
      interpolation_pool[i].check();
    }
    catch(Init_Err) {
      is_init_err = true;
    }

  // check minimization methods
  for (int i = 0; i < minimization_pool.size(); ++i)
    try {
      Method::find_method(minimization_pool[i]);
    }
    catch(Find_Err) {
      std::cout << funame << "method " << minimization_pool[i]
	   << " is not found among the available methods\n";
      is_init_err = true;
    }

  // check optimization method
  if(is_opt)
    try {
      Method::find_method(opt_method);
    }
    catch(Find_Err) {
      std::cout << funame << "optimization method " << opt_method
	   << " is not found among the available methods\n";
      is_init_err = true;
    }

  if (exe_dir[0] != '/') {
    std::cout << funame << "the g98 directory " << exe_dir
	    << " should start with absolute path\n";
    is_init_err = true;
  }

  // throw initializatio error
  if (is_init_err)
    throw Init_Err();

  // set environment
  setenv("GAUSS_SCRDIR", ".", 1);
  setenv("GAUSS_EXEDIR", exe_dir.c_str(), 1);
  setenv("PATH", exe_dir.c_str(), 1);

  // set potential correction
  if(is_clib && is_csym) {
    _ispcorr = true;
    void* handle = dlopen(clib.c_str(), RTLD_NOW);
    const char* messg = dlerror();
    if(messg) {
      std::cout << funame << messg << "\n";
      throw Init_Err();
    }
    _pcorr = (pcorr_t)dlsym(handle, csym.c_str());
    messg = dlerror();
    if(messg) {
      std::cout << funame << messg << "\n";
      throw Init_Err();
    }
  }
}

void Gauss::print_geom ()
{
  Log::out << "Geometry(Bohr):\n\n";
  for (int i = 0; i < cluster.size(); ++i)
    Log::out << cluster[i];
  Log::out << "\n";
}

Gauss::Method* Gauss::min_pot (int flags) throw(Err)
{
  static const char* funame = "Gauss::min_pot: ";
  Method::clean_all();

  if(debug >= Log::INFO)
    print_geom();

  Method* res = 0;
  for (int i = 0; i < minimization_pool.size(); ++i) {
    Method* curr = minimization_pool[i]->apply(flags);
    if (curr &&	(!res || curr->abs_energy() < res->abs_energy()))
      res = curr;
  }

  if (debug >= Log::INFO)
    Method::print_all_info(Log::out);

  if (!res) {
    if(debug >= Log::ERROR)
      Log::out << funame << "energy calculation failed\n"; 
    throw Err();
  }
  if(res->rel_energy() < -1.0) {
    if(debug >= Log::ERROR) {
      Log::out << funame << "energy " << res->rel_energy()
	      << " is too negative, check *." << fail_count << " files\n";
      res->save_log(fail_count++);
    }
    throw Err();
  }

  if(flags & FORCE)
    res->read_forces();

  return res;
}

double Gauss::lin_pot (int flags) throw(Err)
{
  Method::clean_all();

  if(debug >= Log::INFO)
    print_geom();

  double ener;
  for (int i = 0; i < interpolation_pool.size(); ++i)
    try {
      ener = interpolation_pool[i].apply(flags);
      if (debug >= Log::INFO)
	Method::print_all_info(Log::out);
      return ener;
    }
    catch(Err) {}

  if(debug >= Log::INFO)
    Method::print_all_info(Log::out);

  throw Err();
}

