#ifndef MY_GAUSS
#define MY_GAUSS

#include <string>
#include <vector>
#include <fstream>
#include "math.hh"

namespace Gauss {

  using namespace std;

  extern int debug;

  extern string exe_dir;  // directory with gausssian executables
  extern string exe_name; // g98 executable name

  // reactive system specification
  struct Atom;
  extern vector<Atom> cluster;
  extern int multiplicity;
  extern int charge;

  // Error classes
  class Err {};

  // File handling errors
  class File_Err: public Err {};
  class Open_Err: public File_Err{}; // File opening errors
  class Form_Err: public File_Err{}; // Wrong format
  class EOF_Err:  public File_Err{}; // End-of-file encountered

  // Math errors
  class Math_Err: public Err {};
  // Indexing errors
  class Range_Err: public Err {};
  // Search errors
  class Find_Err: public Err {};
  // Initialization errors
  class Init_Err: public Err {};
  // Run errors
  class Run_Err: public Err {}; 

  struct Atom {
    int num;                 // atomic number
    double pos [3];          // atom's position
    double force [3];        // force exerted on the atom
  };

  ostream& operator<< (ostream&, const Atom&);

  // operation flags
  enum {
    FORCE  = 4, // calculate forces flag
    READ   = 2, // guess = read
    DIRECT = 1 // apply method literally (do not use failsafe methods); 
  };

  // buffer of strings associated with a file input stream
  class Buffstream { 

    ifstream from;
    vector<string> buff;

  public:

    Buffstream (const string&) throw (Open_Err);
    const string& operator[] (int) throw (Err);
    void skip (int =1) throw (Err);
    Buffstream& operator>> (double&) throw (File_Err);
    Buffstream& operator>> (string&) throw (File_Err);
    void close() throw () {from.close(); buff.clear(); }

  };

  class Method;

  // method associated with the id
  struct Methp {

    string id;
    Method* mp;

    Methp () : mp(0) {}
    Methp (const string& s) : id(s), mp(0) {}
    Method* operator-> () const throw(Init_Err); 

  };

  istream& operator>> (istream&, Methp&);
  ostream& operator<< (ostream&, const Methp&);

  // search pattern
  class Pattern {

    vector<string> keys; 

  public:

    Pattern (ifstream&) throw(Err);
    bool test (Buffstream&) const throw();

  };

  class Value_pattern: public Pattern {

    double _weight;

  public:

    Value_pattern (ifstream&) throw(Err);
    double weight () const { return _weight; }
    void operator /= (double val) { _weight /= val; }
  };

  class Message_pattern: public Pattern {

    string _message;

  public:

    Message_pattern (ifstream&) throw(Err);
    const string& message () const { return _message; }

  };
    
  class Fail_pattern: public Message_pattern {

    vector<Methp> methods; // fail-safe methods

  public:

    Fail_pattern (ifstream&) throw(Err);
    int size() const { return methods.size(); }
    Methp& method(int i) throw(Range_Err);
  };

  // g98 calculation method class  
  class Method {

  public:
    // status flags
    enum Stat {
      INIT,    // method was not applied yet
      SUCCESS, // calculation succeeded
      FAIL     // calculation failed
    };

  private:

    string id;      // base_name; it should be unique
    string command; // gaussian command string
 
    // methods to calculate the initial guess
    vector<Methp> init_guess_methods;
    bool is_init_guess_min; // is initial guess is a minimum of init. guess
    // methods or just the first which will work
    bool is_init_guess_mix;
    bool is_init_guess_required; // is initial guess required or one can
    // start without checkpoint file

    // energy search patterns.
    vector<Value_pattern> ener_patt;
    // failure patterns and corresponding fail-safe methods
    vector<Fail_pattern> fail_patt;
    // fail patterns to rerun method
    vector<Message_pattern> run_patt;
    // methods to apply in the case of unknown failure
    vector<Methp> default_fail_safe_methods;
    
    int mcorr;     // multiplicity correction (divided by 2)
    double ref_en;   // reference energy

    string memory;

    bool is_checked;

    Stat _status;   // status of method application
    double _abs_en; // result energy
    Method* failsafe; // fail-safe method

    double backup_ener;

    // array of actual methods read from the input file
    static vector<Method*> avail_methods;
    
    // extract energy from fchk file
    void check (vector<string>&) throw(Init_Err); // check for infinite loops

  public:

    Method (ifstream&) throw(Err);
    Method* apply (int =0) throw();
    const string& name   () const { return id; }
    void read_forces (vector<D3>* =0) const throw(Err);
    void save_log (int) const throw(); // save g98 com, log, etc. files
    void print_info (ostream&); // print calculated energies, etc.
    //void clean () throw(); 

    void backup  () throw(Init_Err);
    void restore () throw(Init_Err);
    
    int status () const { return _status; }
    void init() { _status = INIT; }

    double rel_energy  () const throw(Init_Err);
    double abs_energy  () const throw(Init_Err);

    static void find_method (Methp&) throw (Find_Err);
    static void clean_all () throw();
    static int pool_size() { return avail_methods.size(); }
    static void check_all () throw(Init_Err); // check all for infinite loops
    // print geometry, calculated energies, etc. from all methods
    static void print_all_info (ostream&); 
  };

  // linear interpolation of several methods
  class Interpolation {

    vector<Methp> methods;
    vector<double> weights;
    
  public:
    
    Interpolation (ifstream&) throw(Err);
    double apply (int) const throw(Err);
    void check () throw(Init_Err); // check that methods are available 

  };

  extern vector<Interpolation> interpolation_pool; // interpolations to try
  extern vector<Methp>         minimization_pool; // minimization to try

  // initialize Gaussian from the file
  void init (const string&) throw(Err);
  Method* min_pot (int =0) throw(Err); // minimization potential
  double lin_pot (int =0) throw(Err); // interpolation potential

  // correction to the potential 
  double pcorr(const int&, const double [], double []);
  bool  ispcorr ();
  void print_geom ();

  // optimization
  extern Methp opt_method;
}

#endif
